# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# version 0.8.4
# date 1 aug 2023

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT
# **FIN  : Finish this

import numpy as np
from .utils import flatten, cps, GM_Estimator, get_support

class Parameter(object):
    def __init__(self, attributes_dict: dict):
        '''
        value, prior, limits, type, name, prargs, ptformargs,
        fixed, sigma, GM_parameter, posterior
        '''
        self.name = None

        for attr in attributes_dict:
            setattr(self, attr, attributes_dict[attr])

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self.name)


class Parameter_Block(object):
    def __init__(self, params, block_model=None, block_name=None, block_type=None, model_script='', display=True, math_display='', is_iterative=False, parameterisation=None, number=0, moav=0):
        self.list_ = np.array(params)
        self.ndim_ = len(self.list_)

        self.model = block_model
        self.name_ = block_name
        self.type_ = block_type
        self.display_on_data_ = display
        self.number_ = number
        self.bnumber_ = 0
        self.is_iterative_ = is_iterative
        self.slice = slice(None)
        self.additional_priors = []
        self.additional_priors_bool = False
        self.dynamics_bool = False
        self.model_script = model_script

        self.additional_parameters = []

        self.int1 = 1
        self.moav = moav
        self.bool1 = 1
        self.bool2 = 1

        self.C_ = []
        self.A_ = []

        self.gC_ = []
        self.gA_ = []

        empty_lists = ['extra_args', 'A_', 'C_', 'additional_priors']

        for attribute in empty_lists:
            setattr(self, attribute, [])

        self.parameterisation = parameterisation
        self.math_display_ = math_display

        self.b_fixed = self.get_attr('fixed')
        self.notfixed_bool_mask = [f is None for f in self.b_fixed]


    def get_attr(self, call):
        if type(call) == str:
            return [getattr(param, call) for param in self]
        elif type(call) == list:
            return [[getattr(param, c) for param in self] for c in call]


    def set_attr(self, call, val, silent=False):
        x = ''
        for i in range(len(self)):
            setattr(self[i], call, val[i])
            if not silent:
                x += '{0} {1} set to {2}\n'.format(self[i], call, val[i])
        return x


    def add_additional_priors(self, priors):
        for prior in priors:
            self.additional_priors.append(prior)
        self.additional_priors_bool = True


    def add_additional_parameters(self, param_dict):
        self.additional_parameters.append(Parameter(param_dict))


    def refresh_block(self):
        self.C_ = []
        self.A_ = []
        ndim = len(self)
        subscript_nums = ['', '₁', '₂', '₃', '₄', '₅', '₆', '₇', '₈', '₉']

        for i in range(ndim):
            if self[i].fixed is not None:
                self.A_.append(i)
                self[i].value = self[i].fixed
                self[i].prior = 'Fixed'
                self[i].limits = [np.nan, np.nan]
                ndim -= 1
            else:
                self.C_.append(i)

            if self[i].prior == 'Uniform':
                self[i].display_prior = '~𝓤 ({}, {})'.format(*np.round(self[i].limits, 3))
            elif self[i].prior == 'Normal':
                self[i].display_prior = '~𝓝 ({}, {})'.format(*np.round(self[i].prargs, 3))
            elif self[i].prior == 'GaussianMixture':
                gmparam = self[i].GM_parameter
                self[
                    i
                ].display_prior = f'𝛴{subscript_nums[gmparam.n_components]}~𝓝 ({list(np.round(gmparam.means, 3))}, {list(np.round(gmparam.sigmas, 3))})'
            elif self[i].prior == 'Fixed':
                self[i].display_prior = f'~𝛿 (x - {self[i].value})'
            else:
                self[i].display_prior = f'Method not built for {self[i].prior}'

        self.b_fixed = self.get_attr('fixed')
        self.notfixed_bool_mask = [f is None for f in self.b_fixed]

        self.ndim_ = ndim


    def calc_priors(self, theta):
        lp = 0.
        for i in range(len(self)):
            lp += getattr(my_stats, self[i].prior)(theta[self.slice][i], self[i].limits, self[i].prargs)

            #print(self[i].name, self[i].limits, theta[self.slice][i], lp)
            if lp == -np.inf:
                return lp

        for p in self.additional_parameters:
            if p.has_prior:
                if p.name == 'Amplitude':
                    x = theta[self.slice][1]**2 + theta[self.slice][2]**2
                elif p.name == 'Eccentricity':
                    x = theta[self.slice][3]**2 + theta[self.slice][4]**2
                elif p.name == 'Hill':
                    kplan = p.prargs[0]
                    if kplan > 1:
                        x = self.get_PAE(theta, kplan)
                else:
                    continue

                lp += getattr(my_stats, p.prior)(x, p.limits, p.prargs)
                if lp == -np.inf:
                    return lp

        return lp


    def get_PAE(self, theta, kplanets):
        ndim = len(self)
        periods = theta[slice(0, ndim*kplanets, ndim)]

        if self.parameterisation in [0, 2, 3]:
            amps = theta[slice(1, ndim*kplanets, ndim)]
            eccs = theta[slice(3, ndim*kplanets, ndim)]

        if self.parameterisation == 1:
            periods = np.exp(periods)
            amps = theta[slice(1, ndim*kplanets, ndim)] ** 2 + theta[slice(2, ndim*kplanets, ndim)] ** 2
            eccs = theta[slice(3, ndim*kplanets, ndim)] ** 2 + theta[slice(4, ndim*kplanets, ndim)] ** 2

        if self.parameterisation == 3:
            eccs = theta[slice(3, ndim*kplanets, ndim)] ** 2 + theta[slice(4, ndim*kplanets, ndim)] ** 2

        return np.array([periods, amps, eccs])


    def __repr__(self):
        return self.name_+'(%i)' % self.ndim_


    def __len__(self):
        return len(self.list_)


    def __getitem__(self, n):
        return self.list_[n]


    def __str__(self):
        return self.name_


    pass


class ReddModel(object):
    def __init__(self, data, bloques):
        self.bloques = bloques
        self.data = data

        self.x = data['BJD'].values
        self.y = data['RV'].values
        self.yerr = data['eRV'].values

        self.bloques_model = []
        self.bloques_error = []
        self.ndata = len(self.x)

        self.mod_fixed = flatten(self.get_attr_param('fixed'))
        self.A_ = []
        self.C_ = []

        self.notfixed_bool_mask = [f is None for f in self.mod_fixed]
        self.ndim__ = np.sum(self.get_attr_block('ndim_'))
        #self.additional_priors_bool = False
        #self.are_additional_priors = False

        self.model_script_no = 0


    def evaluate_model(self, theta):
        mod0 = np.zeros_like(self.y)
        err20 = self.yerr ** 2

        for a in self.A_:
            theta = np.insert(theta, a, self.mod_fixed[a])

        for b in self.bloques:
            if b.moav:
                b.model.fargs[4] = self.y - mod0

            mod, ferr = b.model(theta[b.slice])
            mod0 += mod
            err20 += ferr

        return mod0, err20  # returns model and errors ** 2


    def evaluate_logprior(self, theta):
        lp = 0.

        for a in self.A_:
            theta = np.insert(theta, a, self.mod_fixed[a])

        for b in self:
            lp += b.calc_priors(theta)
            if lp == -np.inf:
                return lp
        return lp


    def evaluate_ptform(self, theta, ptformargs):
        for a in self.A_:
            theta = np.insert(theta, a, self.mod_fixed[a])
        x = np.array(theta)
        for i in range(len(x)):
            a, b = ptformargs[i]
            x[i] =  a * (2. * x[i] - 1) + b
        return x


    def evaluate_loglikelihood(self, theta):
        model, err2 = self.evaluate_model(theta)
        return -0.5 * np.sum((self.y - model) ** 2 / err2 + np.log(err2)) + np.log(2*np.pi) * self.ndata


    def get_GMEstimates(self, chains):
        for b in self:
            for p in b:
                if p.cpointer is not None:
                    p.GM_parameter = GM_Estimator().estimate(chains[:, p.cpointer],
                                                            p.name, p.unit)
                else:
                    p.GM_parameter = p.value


    def refresh__(self):
        self.bloques_model = []
        self.bloques_ins = []
        self.kplan__ = 0
        self.nins__ = 0

        # update slices and sorts blocks

        nt = 0
        ntt = 0
        bn_ = 1
        for b in self:
            b.bnumber_ = bn_
            if b.type_ == 'Keplerian':
                self.bloques_model.append(b)
                self.kplan__ += 1
            if b.type_ == 'Instrumental':
                self.bloques_ins.append(b)
                self.nins__ += 1

            b.refresh_block()
            b.slice = slice(nt, nt+len(b))
            b.slice_true = slice(ntt, ntt+b.ndim_)
            nt += len(b)
            ntt += b.ndim_
            bn_ += 1
        # updates model's ndim
        self.ndim__ = np.sum(self.get_attr_block('ndim_'))
        for b in self.bloques_model:
            if b.dynamics_bool:
                for extra in b.additional_priors:
                    if extra[0] == 'Hill':
                        extra[3][0] = self.kplan__

        # updates model's A_ & C_
        for i in range(len(self.mod_fixed)):
            if self.mod_fixed[i] != None:
                self.A_.append(i)
            else:
                self.C_.append(i)

        # updates block's gA_ & gC_
        tdims = 0
        ntdims = 0
        for b in self:
            b.gC_ = self.C_[tdims:tdims+b.ndim_]

            b.gA_ = self.A_[ntdims:ntdims+(len(b)-b.ndim_)]

            b.cpointer = np.arange(tdims, tdims+b.ndim_)
            j = 0
            for p in b:
                if p.fixed == None:
                    p.cpointer = b.cpointer[j]
                    j += 1
                else:
                    p.cpointer = None

            tdims += b.ndim_
            ntdims += (len(b)-b.ndim_)

        # updates nsai
        self.cornums = []
        for b in self:
            if b.type_ == 'Instrumental':
                self.cornums.append(b.cornum)

    def get_attr_param(self, call):
        return [b.get_attr(call) for b in self]


    def get_attr_block(self, call):
        return [getattr(b, call) for b in self]


    def display_math(self):
        return ' + '.join(self.get_attr_block('math_display_'))


    def write_model_(self, loc='', tail=''):
        saveloc = f'{loc}/temp/'
        self.data.to_csv(f'{saveloc}temp_data{tail}.csv')

        fname = f'{saveloc}/temp_model_{self.model_script_no}{tail}.py'
        with open(fname, 'w') as f:
            # data
            f.write('''
my_data = pd.read_csv('{}temp_data{}.csv')

'''.format(saveloc, tail))

            # model
            f.write('''
def my_model(theta):
    for a in A_:
        theta = np.insert(theta, a, mod_fixed_[a])

    model0 = np.zeros_like(my_data['BJD'])
    err20 = my_data['eRV'] ** 2

''')
            for b in self:
                if b.type_ == 'Keplerian':
                    f.write(open(get_support(f'models/{b.model_script}')).read().format(b.slice))
                elif b.type_ == 'Instrumental':
                    f.write(open(get_support(f'models/{b.model_script}')).read().format(b.ins_no,
                                                               b.slice,
                                                               b.moav))
                elif b.type_ == 'Acceleration':
                    f.write(open(get_support(f'models/{b.model_script}')).read().format(b.slice))
            f.write('''
    return model0, err20


''')
        self.model_script_no += 1
        return fname


    def __getitem__(self, n):
        return self.bloques[n]


    def __repr__(self):
        return str([b for b in self])


class my_stats():
    def Uniform(x, limits, args):
        if limits[0] <= x <= limits[1]:
            return 0.
        else:
            return -np.inf

    def Normal(x, limits, args):
        if limits[0] <= x <= limits[1]:
            mu, var = args[0], args[1]**2
            return - 0.5 * (np.log(2*np.pi*var) + (x - mu)**2/var)
        else:
            return -np.inf

    def Fixed(x, limits, args):
        return 0.

    def Jeffreys(x, limits, args):
        if limits[0] <= x <= limits[1]:
            return 0.
        else:
            return -np.inf

    def Hill(x, limits, args):
        kplanets = args[0]
        starmass = args[1]
        periods, amps, eccs = x

        gamma = np.sqrt(1 - eccs)
        sma, minmass = cps(periods, amps, eccs, starmass)
        orden = np.argsort(sma)
        sma = sma[orden]  # in AU
        minmass = minmass[orden]  # in Earth Masses

        periods, amps, eccs = periods[orden], amps[orden], eccs[orden]
        M = starmass * 1047.56 + np.sum(minmass)  # jupiter masses
        mu = minmass / M

        for k in range(kplanets-1):
            alpha = mu[k] + mu[k+1]
            delta = np.sqrt(sma[k+1] / sma[k])

            LHS = alpha**-3 * (mu[k] + (mu[k+1] / (delta**2))) * (mu[k] * gamma[k] + mu[k+1] * gamma[k+1] * delta)**2
            RHS = 1 + (3./alpha)**(4./3) * (mu[k] * mu[k+1])
            #LHS = delta
            #RHS = 2.4*alpha**(1./3)
            if LHS > RHS:
                pass
            else:
                return -np.inf

        return 0.

    def GaussianMixture(x, limits, args):
        #  = my_bgm.bgm_estimator[0]
        if limits[0] <= x <= limits[1]:
            return 0
        else:
            return -np.inf

#



# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# version 0.8.4
# date 1 aug 2023

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT
# **FIN  : Finish this

import numpy as np
import pandas as pd

import sys, os
from contextlib import contextmanager

from termcolor import colored
from sklearn.mixture import GaussianMixture


_ROOT = os.path.dirname(__file__)

def get_support(path):
    return os.path.join(_ROOT, 'support', path)


from functools import singledispatch
import shutil

class nuker(object):
    def __init__(self, loc=''):
        self._loc = loc
        self._target_folder = ''
        self._dr = ''
        self.IamSure = False
        self._nuclear = False

    @property
    def dr(self):
        return self._dr

    @dr.setter
    def dr(self, val):
        self._dr = val

    @property
    def loc(self):
        return self._loc

    @loc.setter
    def loc(self, val: str):
        self._loc = val


    def aim(self, target):
        if type(target) == str:
            self._target_folder = target
            self._dr = f'{self._loc}datalogs/{self._target_folder}/'
            self._target_list = os.listdir(self._dr)

        if type(target) == tuple:
            holder = []
            tmin, tmax = target[0], target[1] + 1
            for i in range(tmin, tmax):
                run_name = f'run_{i}'
                if run_name in self._target_list:
                    holder.append(run_name)
            self._target_list = holder

        if type(target) == bool and target:
            self._dr = f'{self._loc}datalogs/'
            self._nuclear = True
            self._target_list = os.listdir(self._dr)

        for x in self._target_list:
            if x[0] == '.':
                self._target_list.remove(x)


    def nuke(self):
        if not self.IamSure:
            print('You are about to delete the following directories:\n')
            for tg in self._target_list:
                print(f'    - {self._dr}{tg}\n')
            print('if you are really sure, set nuker.IamSure = True')
        else:
            if type(self._target_list) == str:
                shutil.rmtree(self._dr)
            if type(self._target_list) == list:
                for tg in self._target_list:
                    shutil.rmtree(f'{self._dr}{tg}')


def fold(x, y, yerr=None, per=None):
    if per is None:
        per = 2. * np.pi
    x_f = x % per
    order = np.argsort(x_f)
    if yerr is None:
        return x_f[order], y[order]
    else:
        return x_f[order], y[order], yerr[order]


def fold_dataframe(df, per=None):
    if per is None:
        per = 2 * np.pi
    df['BJD'] = df['BJD'] % per
    return df.sort_values('BJD')


def minmax(x):
    return np.amin(x), np.amax(x)


def flatten(t):
    return [item for sublist in t for item in sublist]


def cps(pers, amps, eccs, starmass):
    #sma, minmass = np.zeros(kplanets), np.zeros(kplanets)
    G = 6.674e-11  # m3 / (kg * s2)
    #m2au = 6.685e-12  # au
    #kg2sm = 5.03e-31  # solar masses
    #s2d = 1.157e-5  # days
    #G_ = G * m2au**3 / (kg2sm*s2d)  # au3 / (sm * d2)

    consts = 4*np.pi**2/(G*1.99e30)

    sma = ((pers*24*3600)**2 * starmass / consts)**(1./3) / 1.49598e11
    minmass = amps / ( (28.4329/np.sqrt(1. - eccs**2.)) * (starmass**(-0.5)) * (sma**(-0.5)) )

    return sma, minmass


def hill_check(p, a, e, sm=0.33):
    #kp = len(p)

    sma, minmass = cps(p, a, e, sm)
    o = np.argsort(sma)

    sma, minmass = sma[o], minmass[o]
    p, a, e = p[o], a[o], e[o]

    gamma = np.sqrt(1 - e**2)
    LHS, RHS = [], []
    for k in np.arange(len(p)):
        mm = np.array([minmass[k], minmass[k+1]])
        M = sm * 1047.56 + np.sum(mm)
        mu = mm / M
        alpha = np.sum(mu)
        delta = np.sqrt(sma[k+1] / sma[k])
        LHS.append(alpha**-3 * (mu[k] + (mu[k+1] / (delta**2))) * (mu[k] * gamma[k] + mu[k+1] * gamma[k+1] * delta)**2)
        RHS.append(1 + (3./alpha)**(4./3) * mu[k] * mu[k+1])

    return LHS, RHS

'''
def amd(p, a, e, sm=0.33):
    kp = len(p)
    sma, minmass = cps(p, a, e, kp, sm)
    o = np.argsort(sma)

    sma, minmass = sma[o], minmass[o]
    p, a, e = p[o], a[o], e[o]

    mu = G * sm
    eps = np.sum(minmass) / sm
    gamma = minmass[0] / minmass[1]
    lambd = minmass * np.sqrt(mu*sma)
    alpha = sma[0] / sma[1]

    C = gamma * np.sqrt(alpha) * (1 - np.sqrt(1-e[0]**2)*1) + 1 - np.sqrt(1-e[1]**2)*1
    C = gamma * np.sqrt(alpha) + 1
    RHS = gamma * np.sqrt(alpha) + 1 - (1+gamma)**(1.5) * np.sqrt( (alpha/(gamma+alpha)) * (1+ (3**(4./3) * eps**(2./3) * gamma)/(1+gamma)**2))

    LHS = gamma * np.sqrt(alpha) * (1 - np.sqrt(1-e[0]**2)*np.cos(i[0])) + 1 - np.sqrt(1 - e[1]**2)*np.cos(i[1])
'''

def delinearize(x, y):
    A = x**2 + y**2
    B = np.arccos(y / (A ** 0.5)) if A != 0 else 0
    if x < 0:
        B = 2 * np.pi - B
    return np.array([A, B])


def adelinearize(s, c):
    # x sine, y cosine
    A = s**2 + c**2
    B = np.zeros_like(A) if A.all() == 0 else np.arccos(c / (A ** 0.5))
    B[s<0] = 2 * np.pi - B[s<0]

    #where is slower
    #B = np.where(x>0, np.arccos(y / (A ** 0.5)), 2 * np.pi - np.arccos(y[x<0] / (A[x<0] ** 0.5)))
    return np.array([A, B])

'''
def delinearize_pymc3(x, y):
    import aesara_theano_fallback.tensor as tt
    import exoplanet as xo

    A = x**2 + y**2
    B = tt.arccos(y / (A ** 0.5))
    if x < 0:
        B = 2 * np.pi - B
    return tt.array([A, B])
'''

def getExtremePoints(data, typeOfExtreme = None, maxPoints = None):
    """
    from https://towardsdatascience.com/modality-tests-and-kernel-density-estimations-3f349bb9e595
    """
    a = np.diff(data)
    asign = np.sign(a)
    signchange = ((np.roll(asign, 1) - asign) != 0).astype(int)
    idx = np.where(signchange == 1)[0]

    if typeOfExtreme == 'max' and data[idx[0]] < data[idx[1]]:
        idx = idx[1:][::2]

    elif typeOfExtreme == 'min' and data[idx[0]] > data[idx[1]]:
        idx = idx[1:][::2]

    elif typeOfExtreme is not None:
        idx = idx[::2]

    # sort ids by min value
    if 0 in idx:
        idx = np.delete(idx, 0)
    if (len(data) - 1) in idx:
        idx = np.delete(idx, len(data)-1)

    idx = idx[np.argsort(data[idx])]
    # If we have maxpoints we want to make sure the timeseries has a cutpoint
    # in each segment, not all on a small interval
    if maxPoints is not None:
        idx = idx[-maxPoints:]
        if len(idx) < maxPoints:
            return (np.arange(maxPoints) + 1) * (len(data)//(maxPoints + 1))

    return idx


def plot_extreme(data, n=10):
    import matplotlib.pyplot as pl
    x, y = data

    idx = getExtremePoints(y, typeOfExtreme='max')
    pl.plot(x,y)
    ax = pl.gca()
    pl.scatter(x[idx], y[idx], s=40, c='red')

    for i in idx[:n]:
        ax.annotate(f' Max = {np.round(x[i], 2)}', (x[i], y[i]))
    pl.show()


def ensure_dir(name, loc=''):
    dr = f'{loc}datalogs/{name}/run_1'
    while os.path.exists(dr):
        aux = int(dr.split('_')[-1]) + 1
        dr = dr.split('_')[0] + '_' + str(aux)

    os.makedirs(dr)
    os.makedirs(f'{dr}/plots')
    os.makedirs(f'{dr}/plots/histograms')
    os.makedirs(f'{dr}/plots/GMEstimates')
    os.makedirs(f'{dr}/plots/traces')
    os.makedirs(f'{dr}/plots/models')
    os.makedirs(f'{dr}/plots/models/uncertainpy')

    os.makedirs(f'{dr}/samples')
    os.makedirs(f'{dr}/samples/posteriors')
    os.makedirs(f'{dr}/samples/likelihoods')
    os.makedirs(f'{dr}/samples/chains')

    os.makedirs(f'{dr}/temp')
    os.makedirs(f'{dr}/temp/models')

    os.makedirs(f'{dr}/restore')
    os.makedirs(f'{dr}/restore/backends')

    return dr


def set_pos0(setup, model_obj):
    ntemps, nwalkers, nsteps = setup
    ndim = model_obj.ndim__
    pos = np.zeros((ntemps, nwalkers, ndim))

    for t in range(ntemps):
        j = 0
        for b in model_obj:
            for p in b:
                if p.fixed is None:
                    m = (p.limits[1] + p.limits[0]) / 2
                    r = (p.limits[1] - p.limits[0]) / 2
                    dist = np.sort(np.random.uniform(0, 1, nwalkers))

                    if (
                        b.parameterisation != 0 and b.type_ == 'Keplerian'
                    ) and p != b[0]:
                        r *= 0.707

                    pos[t][:, j] = r * (2 * dist - 1) + m
                    np.random.shuffle(pos[t, :, j])
                    j += 1
    return list(pos)


def pos0_tested(setup, model_obj, max_repeats=100):
    p0 = set_pos0(setup, model_obj)

    ntemps, nwalkers, nsteps = setup

    is_bad_position = True
    repeat_number = 0
    while is_bad_position and repeat_number < max_repeats:
        is_bad_position = False
        for t in range(ntemps):
            for n in range(nwalkers):
                position_evaluated = p0[t][n]
                if model_obj.evaluate_logprior(position_evaluated) == -np.inf:
                    is_bad_position = True
                    p0[t][n] = set_pos0(setup, model_obj)[t][n]
        repeat_number += 1
    return p0


@contextmanager
def nullify_output(suppress_stdout=True, suppress_stderr=True, suppress_stdin=True):
    stdout = sys.stdout
    stderr = sys.stderr
    stdin = sys.stdin
    devnull = open(os.devnull, "w")
    try:
        if suppress_stdout:
            sys.stdout = devnull
        if suppress_stderr:
            sys.stderr = devnull
        if suppress_stdin:
            sys.stdin = devnull
        yield
    finally:
        if suppress_stdout:
            sys.stdout = stdout
        if suppress_stderr:
            sys.stderr = stderr
        if suppress_stdin:
            sys.stdin = stdin


class importer:
	def	__init__(self):
		self.c = 1
	def add_path(self, x):
		sys.path.insert(self.c, x)
		self.c += 1
	def restore_priority(self):
		self.c = 1


class binner:
    def __init__(self, x, y=None):
        pass
    pass


class reddlog(object):
    def __init__(self):
        self.log = ''
        try:
            self.terminal_width = os.get_terminal_size().columns
        except:
            self.terminal_width = pd.get_option('display.width')
        self.baddies_list = ["\x1b[", '0m', '1m', '4m', '7m', '31m', '32m']

    def saveto(self, location):
        np.savetxt(f'{location}/log.dat', np.array([self.log]), fmt='%100s')

    def help(self):
        print('Colors: grey, red, green, yellow, blue, magenta, cyan, white')
        print('On_Colors: on_<color>')
        print('Attrs: bold, dark, underline, blink, reverse, concealed')


    def __call__(self, msg, center=False, save=True, c=None, oc=None, attrs=None):
        if attrs is None:
            attrs = []
        if save:
            msg0 = msg
            for b in self.baddies_list:
                msg0 = msg0.replace(b, '')

            self.log += msg0
        if center:
            msg = msg.center(self.terminal_width)
        if c:
            msg = colored(msg, c, oc, attrs)
        print(msg)

'''
class KDE_estimator:
    def __init__(self):
        self.kde_estimator = []
        self.bandwidths = np.arange(0.1, 1.5, .1)
        self.train_length = 5000
        self.chain_length = 5000
        self.x_test = []

    def estimate(self, chain_p, p=0, thin_by=1):
        t_start = time.time()
        x_train = np.array([[c] for c in chain_p[::thin_by]])
        self.x_test.append(np.linspace(np.amin(x_train), np.amax(x_train), self.train_length)[:, np.newaxis])

        kde = KernelDensity(kernel='gaussian')
        grid = GridSearchCV(kde, {'bandwidth': self.bandwidths})

        grid.fit(x_train)
        self.kde_estimator.append(grid.best_estimator_)
        t_end = time.time()
        print(f'Time to estimate parameter {p} is {t_end - t_start} seconds')

    def auto_estimate(self, chain):
        nsteps, ndim = chain.shape
        thin_by_auto = nsteps // self.train_length
        for i in range(ndim):
            self.estimate(chain[:, i], p=i, thin_by=thin_by_auto)
'''

class GaussianMixture_addon(GaussianMixture):
    def __init__(self, name, unit, **kw):
        # this one is per parameter
        self.name = name
        self.unit = unit
        if self.unit is None:
            self.unit = ''
        self.means = None
        self.covariances = None
        self.sigmas = None
        self.weights = None
        self.mixture_mean = None
        self.mixture_variance = None
        self.mixture_sigma = None
        super().__init__(**kw)

    def update(self):
        self.means = self.means_.flatten()
        self.covariances = self.covariances_.flatten()
        self.sigmas = np.sqrt(self.covariances)
        self.weights = self.weights_
        self.mixture_mean = self.Mixture_Mean()
        self.mixture_variance = self.Mixture_Variance()
        self.mixture_sigma = np.sqrt(self.mixture_variance)

    def Mixture_Variance(self):
        s1 = np.sum(self.weights*self.covariances)
        s2 = np.sum(self.weights*self.means**2)
        s3 = np.sum(self.weights*self.means)**2
        return s1 + s2 - s3

    def Mixture_Mean(self):
        return np.sum(self.means*self.weights)


    def __repr__(self):
        return f'{self.means}; {self.sigmas}'


class GM_Estimator:
    def __init__(self):
        # This one is per param
        self.gm_estimator = None
        self.max_n_components = 5
        self.BIC0_ = np.inf
        self.BIC_Tolerance = 0
        self.name__ = 'GM_Estimator'


    def estimate(self, chain_p, name, unit):
        comp_bic = self.BIC0_
        gm0 = None
        chain_p = chain_p[:, np.newaxis]

        for n in range(self.max_n_components):
            gm = GaussianMixture_addon(name, unit, **{'n_components':n+1}).fit(chain_p)
            gm.update()
            mu = gm.means_
            sel_bic = gm.bic(mu)
            if sel_bic - comp_bic < self.BIC_Tolerance:
                comp_bic = sel_bic
                gm0 = gm

        self.gm_estimator = gm0
        return self.gm_estimator


class DataWrapper(object):
    def __init__(self, target_name, read_loc=''):
        self.target_name = target_name
        self.RV_PATH = f'{read_loc}datafiles/{self.target_name}/RV/'

        empty_lists = ['ndata', 'ncols', 'nsai', 'data', 'RV_labels',
                       'RV_sets', 'PM_sets']
        for attribute in empty_lists:
            setattr(self, attribute, [])


    def add_data__(self, filename):
        data = np.loadtxt('{0}{1}'.format(self.RV_PATH, filename))

        ndat, ncol = data.shape

        self.ndata.append(ndat)
        self.ncols.append(ncol)
        self.RV_labels.append(filename)

        names = ['BJD', 'RV', 'eRV']

        # identify and name SAI
        nsa = ncol - 3
        if nsa > 0:
            names.extend(f'Staract {len(self.ndata)} {j}' for j in range(nsa))
        self.nsai.append(nsa)

        df = pd.DataFrame(data, columns=names)

        # substract RV
        if abs(df.mean()['RV']) > 1e-6:
            df['RV'] -= df.mean()['RV']

        for nam in names:
            if nam[:3] == 'Sta':
                if abs(df.mean()[nam]) > 1e-6:
                    df[nam] -= df.mean()[nam]
                    df[nam] = df[nam] / (df.max()[nam] - df.min()[nam])
        # create another column containing flags for the instrument
        df.insert(loc=3, column='Flag', value=np.ones(ndat, int) * len(self.ndata))
        self.data.append(df)

        return 'Reading data from {0}'.format(filename)


    def add_all__(self):
        my_files = list(np.sort(os.listdir(self.RV_PATH)))
        # mac os fix
        try:
            my_files.remove('.DS_Store')
        except:
            pass
        x = ''.join('\n'+self.add_data__(file) for file in my_files)
        return x+'\n'


    def get_data__(self, sortby='BJD'):
        return pd.concat(self.data).sort_values(sortby)


    def get_metadata__(self):
        return [getattr(self, attribute) for attribute in ['ndata', 'ncols', 'nsai', 'RV_labels']]


    def get_data_raw(self, sortby):
        holder = pd.concat(self.data).sort_values(sortby)
        x, y, yerr = holder.BJD, holder.RV, holder.eRV
        return [x, y, yerr]


class ModelWrapper(object):
    def __init__(self, func_model, fargs=None, fkwargs=None):
        if fargs is None:
            fargs = []
        if fkwargs is None:
            fkwargs = {}
        self.func = func_model
        self.fargs = fargs
        self.fkwargs = fkwargs

    def __call__(self, x):
        return self.func(x, *self.fargs, **self.fkwargs)


#



# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# version 0.8.4
# date 1 aug 2023

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT
# **FIN  : Finish this

import numpy as np
import kepler

def norm(x, mu, sig):
    izq = 1 / (sig*np.sqrt(2*np.pi))
    der = np.exp(-0.5*((x-mu)/sig)**2)

    return izq*der


def Line_Model(theta, *args, **kwargs):
    data = args[0]
    x, y, yerr = data

    m, b, logf = theta
    return m * x + b


def DG_Model(theta, *args, **kwargs):
    data = args[0]
    x, y, yerr = data
    mu1, mu2, sigma1, sigma2, logf = theta

    return norm(x, mu1, sigma1) + norm(x, mu2, sigma2), 0


def Keplerian_Model(theta, *args, **kwargs):
    # **EVAL, data vs just time
    per, A, phase, ecc, w = theta

    freq = 2. * np.pi / per  # freq space
    M = freq * args[0] + phase  # mean anomaly
    E = np.array([kepler.solve(m, ecc) for m in M])  # eccentric anomaly
    f = (np.arctan(((1. + ecc) ** 0.5 / (1. - ecc) ** 0.5) * np.tan(E / 2.)) * 2.)  # true anomaly
    model = A * (np.cos(f + w) + ecc * np.cos(w))

    return model, 0


def Keplerian_Model_1(theta, *args, **kwargs):
    P, As, Ac, S, C = theta

    per = np.exp(P)
    A = As ** 2 + Ac ** 2
    ecc = S ** 2 + C ** 2

    if ecc < 1e-5:
        w = 0
    else:
        w = np.arccos(C / (ecc ** 0.5))  # longitude of periastron
        if S < 0:
            w = 2 * np.pi - np.arccos(C / (ecc ** 0.5))

    if ecc > 1 or ecc < 0:
        print('\n HER HERE\n')
        print(theta)

    phase = np.arccos(Ac / (A ** 0.5))
    if As < 0:
        phase = 2 * np.pi - np.arccos(Ac / (A ** 0.5))


    freq = 2. * np.pi / per
    M = freq * args[0] + phase
    E = np.array([kepler.solve(m, ecc) for m in M])
    f = (np.arctan(((1. + ecc) ** 0.5 / (1. - ecc) ** 0.5) * np.tan(E / 2.)) * 2.)  # true anomaly

    model = A * (np.cos(f + w) + ecc * np.cos(w))

    return model, 0


def Keplerian_Model_2(theta, *args, **kwargs):
    #model = 0
    per, A, tp, ecc, w = theta
    freq = 2. * np.pi / per

    M = freq * (args[0] - tp)
    E = np.array([kepler.solve(m, ecc) for m in M])
    f = (np.arctan(((1. + ecc) ** 0.5 / (1. - ecc) ** 0.5) * np.tan(E / 2.)) * 2.)  # true anomaly
    model = A * (np.cos(f + w) + ecc * np.cos(w))

    return  model, 0


def Keplerian_Model_3(theta, *args, **kwargs):
    per, A, tp, S, C = theta

    ecc = S ** 2 + C ** 2

    if ecc < 1e-5:
        w = 0
    else:
        w = np.arccos(C / (ecc ** 0.5))  # longitude of periastron
        if S < 0:
            w = 2 * np.pi - np.arccos(C / (ecc ** 0.5))

    freq = 2. * np.pi / per
    M = freq * (args[0] - tp)
    E = np.array([kepler.solve(m, ecc) for m in M])
    f = (np.arctan(((1. + ecc) ** 0.5 / (1. - ecc) ** 0.5) * np.tan(E / 2.)) * 2.)  # true anomaly
    model = A * (np.cos(f + w) + ecc * np.cos(w))

    return  model, 0


def Instrument_Model(theta, *args, **kwargs):
    ins_no = args[0]
    flags = args[1]

    my_mask = (flags == ins_no)

    mod = theta[0] * my_mask  # OFFSET

    new_err = my_mask * theta[1] ** 2

    return mod, new_err


def Instrument_Moav_Model(theta, *args, **kwargs):
    #offset, jitter = theta

    ins_no = args[0]
    flags = args[1]
    maorder = args[2]

    time = args[3]
    residuals = args[4]

    my_mask = (flags == ins_no)

    mod = theta[0] * my_mask  # OFFSET

    if maorder > 0:
        theta_ma = theta[2:]
        t_ = time[my_mask]
        for i in range(len(t_)):
            for c in range(maorder):
                if i > c:
                    dt = abs(t_[i] - t_[i - 1 - c])
                    macoef = theta_ma[2 * c]
                    matime = theta_ma[2 * c + 1]
                    MA = macoef * np.exp(-dt / matime) * residuals[my_mask][i - 1 - c]
                    mod[my_mask][i] += MA
                    residuals[my_mask][i] -= MA


    new_err = my_mask * theta[1] ** 2

    return mod, new_err


def Instrument_Moav_SA_Model(theta, *args, **kwargs):
    # time and residuals should exclusively be the ones for this instrument
    ins_no = args[0]
    flags = args[1]
    maorder = args[2]
    time = args[3]

    staracts = args[4]
    cornum = args[5]

    #print('in model 2', offset*[flags == ins_no])

    my_mask = (flags == ins_no)
    mod = theta[0] * my_mask  # OFFSET
    #residuals = rv - mod

    if cornum:
        for j in range(cornum):
            mod[my_mask] += theta[2 * (maorder + 1) + j] * staracts[j]

    if maorder > 0:
        residuals = modargs[0][-1]
        theta_ma = theta[2:]
        t_ = time[my_mask]
        for i in range(len(t_)):
            for c in range(maorder):
                if i > c:
                    dt = abs(t_[i] - t_[i - 1 - c])
                    macoef = theta_ma[2 * c]
                    matime = theta_ma[2 * c + 1]
                    MA = macoef * np.exp(-dt / matime) * residuals[my_mask][i - 1 - c]
                    mod[my_mask][i] += MA
                    residuals[my_mask][i] -= MA

    new_err = my_mask * theta[1] ** 2

    return mod, new_err


def Instrument_Model_SA(theta, *args, **kwargs):
    # time and residuals should exclusively be the ones for this instrument
    ins_no = args[0]
    flags = args[1]
    maorder = args[2]
    time = args[3]

    staracts = args[4]
    cornum = args[5]

    #print('in model 2', offset*[flags == ins_no])

    my_mask = (flags == ins_no)
    mod = theta[0] * my_mask  # OFFSET
    #residuals = rv - mod

    if cornum:
        for j in range(cornum):
            mod[my_mask] += theta[2 * (maorder + 1) + j] * staracts[j]

    new_err = my_mask * theta[1] ** 2

    return mod, new_err


def Acceleration_Model(theta, *args, **kwargs):
    time = args[0]
    mod = np.polyval(np.concatenate([theta, [0]]), time-time[0])

    return mod, 0


def Empty_Model(theta, *args, **kwargs):
    return 0, 0


'''
# Notes on Keplerian Models

## Model 0 or vanilla
The easiest of them all, it uses
Period, Amplitude, Phase, Eccentricity, Longitude_Periastron

## Model 1 or hou
According to Hou's paper <<LINK>>
Mind that I use log Period here
lPeriod, Amp_sin, Amp_cos, Ecc_sin, Ecc_cos

## Model 2 or Tp
Pretty similar to vanilla
Period, Amplitude, Time_Periastron, Eccentricity, Longitude_Periastron

## Model 3 or Tp Hou
Combination of Time of periastron, circularising the eccentricity.

Period, Amplitude, Phase, Ecc_sin, Ecc_cos

'''










#
